# -*- coding: utf-8 -*-

from django.conf import settings


def data(request):
    d = {}
    d['navsy_debug'] = settings.DEBUG
    if request:
        d['navsy_host'] = '%s://%s' % (request.scheme, request.get_host(), )
        d.update(getattr(request, 'navsy_data', {}))
    return d



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2017-11-22 11:11
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('navsy', '0002_route_published'),
    ]

    operations = [
        migrations.AddField(
            model_name='page',
            name='sitemap_changefreq',
            field=models.CharField(blank=True, choices=[
                    ('always', 'always'),
                    ('hourly', 'hourly'),
                    ('daily', 'daily'),
                    ('weekly', 'weekly'),
                    ('monthly', 'monthly'),
                    ('yearly', 'yearly'),
                    ('never', 'never')
                ], default='never', max_length=50),
        ),
        migrations.AddField(
            model_name='page',
            name='sitemap_item',
            field=models.BooleanField(default=True),
        ),
        migrations.AddField(
            model_name='page',
            name='sitemap_priority',
            field=models.CharField(blank=True, choices=[
                    ('0.0', '0.0'),
                    ('0.1', '0.1'),
                    ('0.2', '0.2'),
                    ('0.3', '0.3'),
                    ('0.4', '0.4'),
                    ('0.5', '0.5'),
                    ('0.6', '0.6'),
                    ('0.7', '0.7'),
                    ('0.8', '0.8'),
                    ('0.9', '0.9'),
                    ('1.0', '1.0')
                ], default='0.5', max_length=5),
        ),
    ]



# -*- coding: utf-8 -*-

from django.conf import settings
from django.http import Http404, HttpResponse, HttpResponseRedirect
from django.shortcuts import render

from navsy import cache
from navsy.utils import import_function, path_utils, pattern_utils, url_utils


def router(request, path=None):

    # force redirect based on settings.APPEND_SLASH value
    path_fixed = path_utils.fix_path(path)

    if path and path != path_fixed:
        fixed_append_slash = path.endswith('/') != path_fixed.endswith('/')
        fixed_double_slash = '//' in path and '//' not in path_fixed
        if fixed_append_slash or fixed_double_slash:
            return HttpResponseRedirect(url_utils.reverse_url(path_fixed))

    path = path_fixed
    # print(path)

    routes_list = cache.get_routes_list()
    # routes_list = list(reversed(routes_list))
    route_not_found = Http404

    for route_data in routes_list:
        route_re = pattern_utils.get_regex_re(
            route_data['page_path'], route_data['regex'])
        route_match = route_re.match(path)
        # print('route match: %s -> %s == %s' % (
        #     True if route_match else False, path, route_re.pattern, ))

        if route_match:
            route_match_kwargs = route_match.groupdict()
            route_match_keys = list(route_match_kwargs.keys())
            route_match_keys.sort(key=lambda arg: route_re.pattern.find(arg))
            route_match_args = [route_match_kwargs[key]
                                for key in route_match_keys]
            # print(route_match_kwargs)
            # print(route_match_keys)
            # print(route_match_args)

            route_data = cache.get_route_data(route_data['pk'])
            # cache.print_data(route_data)
            if route_data.get('published', False):

                page_data = route_data.get('page', { 'published':True })
                # cache.print_data(page_data)
                if page_data.get('published', False):

                    route_obj = route_data['object']

                    try:
                        route_response = __route_response(
                            request,
                            route_obj,
                            *route_match_args,
                            **route_match_kwargs)

                        return route_response

                    except Http404 as route_not_found:
                        continue

    raise route_not_found


def __route_response(request, instance, *args, **kwargs):

    route_obj = instance
    route_data = cache.get_route_data(route_obj.pk)
    # cache.print_data(route_data)

    redirect_url = route_obj.get_redirect_url()
    if redirect_url:
        return HttpResponseRedirect(redirect_url)

    data = {
        'navsy_page': route_obj.page,
        'navsy_route': route_obj,
        'navsy_route_args': args,
        'navsy_route_kwargs': kwargs,
    }

    request.navsy_data = data

    if route_obj.view_template_path:
        return render(request, route_obj.view_template_path)

    if route_obj.view_function_path:
        view_function = import_function(route_obj.view_function_path)
        if view_function:
            if kwargs:
                view_response = view_function(request, **kwargs)
            elif args:
                view_response = view_function(request, *args)
            else:
                view_response = view_function(request)

            if isinstance(view_response, HttpResponse):
                return view_response
            # TODO: return automatic response based on view_response type
            # elif view_response is dict or array, convert it to json and return it
            # elif view_response is bool or number, convert it to string and return it
            else:
                raise Exception('View function "%s" doesn\'t return \
                    an HttpResponse object.' % (route_obj.view_function_path, ))
        else:
            # this should never happen since there is form validation
            raise Exception('View function "%s" is not \
                a valid function.' % (route_obj.view_function_path, ))

    if settings.DEBUG:
        return render(request, 'navsy/default.html')
    else:
        return HttpResponse('')



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.conf import settings
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver
from django.utils.encoding import python_2_unicode_compatible, force_text


@python_2_unicode_compatible
class NodeModel(models.Model):

    parent = models.ForeignKey(
        'self', related_name='children', on_delete=models.CASCADE,
        blank=True, null=True)

    parents_pks = models.CharField(
        max_length=500, blank=True, default='', editable=False)
    parents_count = models.PositiveSmallIntegerField(default=0, editable=False)

    children_pks = models.CharField(
        max_length=500, blank=True, default='', editable=False)
    children_count = models.PositiveSmallIntegerField(
        default=0, editable=False)

    siblings_pks = models.CharField(
        max_length=500, blank=True, default='', editable=False)
    siblings_count = models.PositiveSmallIntegerField(
        default=0, editable=False)

    priority = models.PositiveSmallIntegerField(
        default=0, validators=[MinValueValidator(0), MaxValueValidator(9999)])
    priorities = models.CharField(
        max_length=100, blank=True, default='', editable=False)

    def get_parents(self):
        return list(self.__str_to_qs(self.parents_pks))

    def get_parents_pks(self):
        return self.__str_to_list(self.parents_pks)

    def get_children(self):
        return list(self.__str_to_qs(self.children_pks))

    def get_children_pks(self):
        return self.__str_to_list(self.children_pks)

    def get_siblings(self):
        return list(self.__str_to_qs(self.siblings_pks))

    def get_siblings_pks(self):
        return self.__str_to_list(self.siblings_pks)

    def __list_to_str(self, l):
        s = '|'.join([str(v) for v in l])
        return s

    def __str_to_list(self, s):
        l = [int(v) for v in s.split('|') if v]
        return l

    def __str_to_qs(self, s):
        pks = self.__str_to_list(s)
        qs = self.__class__.objects.filter(pk__in=pks).order_by('-priorities')
        return qs

    def __get_node_data(self, pk_max=0):

        data = {}

        # retrieve parents
        parent_obj = self.parent
        parents_list = []
        obj = parent_obj

        while obj:
            parents_list.insert(0, obj)
            obj = obj.parent

        parents_pks = [obj.pk for obj in parents_list]

        # update parents
        data['parents_pks'] = self.__list_to_str(parents_pks)
        data['parents_count'] = len(parents_list)

        # update children
        children_list = list(self.__class__.objects.filter(
            parent=self.pk).values_list('pk', flat=True))

        data['children_pks'] = self.__list_to_str(children_list)
        data['children_count'] = len(children_list)

        # update siblings
        siblings_list = list(self.__class__.objects.filter(
            parent=parent_obj).exclude(
            pk=self.pk).values_list('pk', flat=True))

        data['siblings_pks'] = self.__list_to_str(siblings_list)
        data['siblings_count'] = len(siblings_list)

        # update priorities
        priorities_objs = list(parents_list) + [self]
        priorities_list = [
            str(min(obj.priority, 9999)).zfill(4) +
            str(max(0, 99999 - obj.pk)).zfill(5) for obj in priorities_objs]
        priorities_list += ['9' * 9] * (10 - len(priorities_list))
        data['priorities'] = self.__list_to_str(priorities_list)

        return data

    @staticmethod
    @receiver(post_delete, dispatch_uid='post_delete_node')
    @receiver(post_save, dispatch_uid='post_save_node')
    def __update_nodes_data(sender, instance, **kwargs):

        if settings.DEBUG:
            import timeit
            start_time = timeit.default_timer()

        if not isinstance(instance, NodeModel):
            return

        objs_manager = sender.objects
        objs_filter = objs_manager.filter
        objs_qs = objs_manager.select_related(
            'parent').prefetch_related('children')
        objs_list = list(objs_qs)

        for obj in objs_list:
            obj_data = obj.__get_node_data()
            objs_filter(pk=obj.pk).update(**obj_data)

        if settings.DEBUG:
            elapsed = timeit.default_timer() - start_time
            print('navsy.tree.models.NodeModel -> update_nodes_data in %ss' % (elapsed, ))

    class Meta:
        abstract = True

    def __str__(self):
        return force_text(self.pk)



# -*- coding: utf-8 -*-

from django.utils.module_loading import import_string

import re


def import_function(path):
    try:
        func = import_string(path)
        if hasattr(func, '__call__'):
            return func
        else:
            return None
    except ImportError:
        return None


def split_words(s):
    return list(filter(bool, re.split(' ,;|', s)))


# from importlib import import_module


# def import_function(path):

#     if not path:
#         return None

#     func_path = path
#     mod_name, func_name = func_path.rsplit('.', 1)

#     if mod_name and func_name:

#         try:
#             mod = import_module(mod_name)

#             try:
#                 func = getattr(mod, func_name)

#                 if hasattr(func, '__call__'):
#                     return func
#                 else:
#                     return None

#             except AttributeError:
#                 return None

#         except ImportError:
#             return None

#     return None
